@@ -59,7 +59,8 @@ upper_lower (HTMLObject *obj, HTMLEngine *e, gpointer data)
 		old_text = HTML_TEXT (obj)->text;
 		HTML_TEXT (obj)->text = up ? g_utf8_strup (old_text, -1) : g_utf8_strdown (old_text, -1);
 		g_free (old_text);
-		HTML_TEXT (obj)->text_bytes = g_utf8_strlen (HTML_TEXT (obj)->text, -1);
+		HTML_TEXT (obj)->text_len = g_utf8_strlen (HTML_TEXT (obj)->text, -1);
+		HTML_TEXT (obj)->text_bytes = strlen (HTML_TEXT (obj)->text);
 	}
 }
 
