@@ -195,11 +195,11 @@ html_engine_save_output_string (HTMLEngineSaveState *state,
 }
 
 gboolean
-html_engine_save_output_buffer (HTMLEngineSaveState *state, const gchar *buffer, int len)
+html_engine_save_output_buffer (HTMLEngineSaveState *state, const gchar *buffer, int bytes)
 {
-	if (len == -1)
-		len = strlen (buffer);
-	return state->receiver (state->engine, buffer, len, state->user_data);
+	if (bytes == -1)
+		bytes = strlen (buffer);
+	return state->receiver (state->engine, buffer, bytes, state->user_data);
 }
 
 
@@ -330,7 +330,6 @@ html_engine_save (HTMLEngine *engine,
 	state.inline_frames = FALSE;
 	state.user_data = user_data;
 	state.last_level = 0;
-	state.nb_intervals = NULL;
 
 	if (! write_header (&state))
 		return FALSE;
@@ -366,7 +365,6 @@ html_engine_save_plain (HTMLEngine *engine,
 	state.inline_frames = FALSE;
 	state.user_data = user_data;
 	state.last_level = 0;
-	state.nb_intervals = NULL;
 
 	/* FIXME don't hardcode the length */
 	html_object_save_plain (engine->clue, &state, 72);
@@ -397,11 +395,6 @@ html_engine_save_buffer_free (HTMLEngineSaveState *state)
 
 	g_string_free (string, TRUE);
 
-	if (state->nb_intervals) {
-		g_slist_free (state->nb_intervals);
-		state->nb_intervals = NULL;
-	}
-	
 	g_free (state);
 }
 
@@ -417,7 +410,7 @@ html_engine_save_buffer_peek_text (HTMLEngineSaveState *state)
 }
 
 int
-html_engine_save_buffer_peek_text_len (HTMLEngineSaveState *state)
+html_engine_save_buffer_peek_text_bytes (HTMLEngineSaveState *state)
 {
 	GString *string;
 	
@@ -427,33 +420,6 @@ html_engine_save_buffer_peek_text_len (HTMLEngineSaveState *state)
 	return string->len;
 }
 
-void
-html_engine_save_add_nb_interval (HTMLEngineSaveState *state, int start_offset, int end_offset)
-{
-	state->nb_intervals = g_slist_prepend (state->nb_intervals, GINT_TO_POINTER (end_offset));
-	state->nb_intervals = g_slist_prepend (state->nb_intervals, GINT_TO_POINTER (start_offset));
-}
-
-void
-html_engine_save_buffer_clear_line_breaks (HTMLEngineSaveState *state, PangoLogAttr *attrs)
-{
-	if (state->nb_intervals) {
-		GSList *cur;
-
-		for (cur = state->nb_intervals; cur; cur = cur->next->next) {
-			int i, start_offset, end_offset;
-			g_return_if_fail (cur->next);
-
-			start_offset = GPOINTER_TO_INT (cur->data);
-			end_offset = GPOINTER_TO_INT (cur->next->data);
-
-			for (i = start_offset; i < end_offset; i ++)
-				if (i > 0 && !attrs [i - 1].is_white)
-					attrs [i].is_line_break = 0;
-		}
-	}
-}
-
 HTMLEngineSaveState *
 html_engine_save_buffer_new (HTMLEngine *engine, gboolean inline_frames)
 {
@@ -467,7 +433,6 @@ html_engine_save_buffer_new (HTMLEngine *engine, gboolean inline_frames)
 		state->inline_frames = inline_frames;
 		state->user_data = (gpointer) g_string_new ("");
 		state->last_level = 0;
-		state->nb_intervals = NULL;
 	}
 
 	return state;
