@@ -88,7 +88,7 @@ DEFINE_UNIMPLEMENTED (set_pen)
 DEFINE_UNIMPLEMENTED (get_black)
 DEFINE_UNIMPLEMENTED (draw_line)
 DEFINE_UNIMPLEMENTED (draw_rect)
-DEFINE_UNIMPLEMENTED (draw_text)
+DEFINE_UNIMPLEMENTED (draw_glyphs)
 DEFINE_UNIMPLEMENTED (draw_spell_error)
 DEFINE_UNIMPLEMENTED (fill_rect)
 DEFINE_UNIMPLEMENTED (draw_pixmap)
@@ -96,7 +96,7 @@ DEFINE_UNIMPLEMENTED (draw_ellipse)
 DEFINE_UNIMPLEMENTED (clear)
 DEFINE_UNIMPLEMENTED (set_background_color)
 DEFINE_UNIMPLEMENTED (draw_shade_line)
-DEFINE_UNIMPLEMENTED (draw_panel)
+DEFINE_UNIMPLEMENTED (draw_border)
 
 DEFINE_UNIMPLEMENTED (set_clip_rectangle)
 DEFINE_UNIMPLEMENTED (draw_background)
@@ -116,6 +116,7 @@ html_painter_init (GObject *object, HTMLPainterClass *real_klass)
 	painter->font_style = GTK_HTML_FONT_STYLE_DEFAULT;
 	painter->font_face = NULL;
 	painter->widget = NULL;
+	painter->clip_width = painter->clip_height = 0;
 }
 
 static void
@@ -134,7 +135,7 @@ text_width (HTMLPainter *painter, PangoFontDescription *desc, const gchar *text,
 	GList *glyphs;
 	gint width = 0;
 
-	pi = html_painter_text_itemize_and_prepare_glyphs (HTML_PAINTER (painter), desc, text, bytes, &glyphs, NULL);
+	pi = html_painter_text_itemize_and_prepare_glyphs (painter, desc, text, bytes, &glyphs, NULL);
 
 	if (pi && glyphs) {
 		GList *list;
@@ -155,7 +156,7 @@ text_width (HTMLPainter *painter, PangoFontDescription *desc, const gchar *text,
 
 static void
 text_size (HTMLPainter *painter, PangoFontDescription *desc, const gchar *text, gint bytes,
-	   HTMLTextPangoInfo *pi, PangoAttrList *attrs, GList *glyphs, gint start_byte_offset,
+	   HTMLTextPangoInfo *pi, GList *glyphs,
 	   gint *width_out, gint *ascent_out, gint *descent_out)
 {
 	gboolean temp_pi = FALSE;
@@ -164,7 +165,7 @@ text_size (HTMLPainter *painter, PangoFontDescription *desc, const gchar *text,
 	gint width = 0;
 	
 	if (!pi) {
-		pi = html_painter_text_itemize_and_prepare_glyphs (painter, desc, text, bytes, &glyphs, attrs);
+		pi = html_painter_text_itemize_and_prepare_glyphs (painter, desc, text, bytes, &glyphs, NULL);
 		temp_pi = TRUE;
 	}
 
@@ -181,7 +182,7 @@ text_size (HTMLPainter *painter, PangoFontDescription *desc, const gchar *text,
 			str = (PangoGlyphString *) gl->data;
 			gl = gl->next;
 			ii = GPOINTER_TO_INT (gl->data);
-			item = pi->entries [ii].item;
+			item = pi->entries [ii].glyph_item.item;
 			pango_glyph_string_extents (str, item->analysis.font, NULL, &log_rect);
 			width += log_rect.width;
 
@@ -216,54 +217,6 @@ text_size (HTMLPainter *painter, PangoFontDescription *desc, const gchar *text,
 	}
 }
 
-static HTMLFont *
-html_painter_real_alloc_font (HTMLPainter *painter, gchar *face, gdouble size, gboolean points, GtkHTMLFontStyle style)
-{
-	PangoFontDescription *desc = NULL;
-	gint space_width, space_asc, space_dsc;
-
-	if (face) {
-		desc = pango_font_description_from_string (face);
-		pango_font_description_set_size (desc, (gint) size);
-	}
-
-	if (!desc || !pango_font_description_get_family (desc)) {
-		if (desc)
-			pango_font_description_free (desc);
-
-		desc = pango_font_description_copy (gtk_widget_get_style (painter->widget)->font_desc);
-	}
-
-	pango_font_description_set_size (desc, size);
-	pango_font_description_set_style (desc, style & GTK_HTML_FONT_STYLE_ITALIC ? PANGO_STYLE_ITALIC : PANGO_STYLE_NORMAL);
-	pango_font_description_set_weight (desc, style & GTK_HTML_FONT_STYLE_BOLD ? PANGO_WEIGHT_BOLD : PANGO_WEIGHT_NORMAL);
-
-	text_size (painter, desc, " ", 1, NULL, NULL, NULL, 0, &space_width, &space_asc, &space_dsc);
-
-	return html_font_new (desc,
-			      space_width,
-			      space_asc, space_dsc,
-			      text_width (painter, desc, "\xc2\xa0", 2),
-			      text_width (painter, desc, "\t", 1),
-			      text_width (painter, desc, "e", 1),
-			      text_width (painter, desc, HTML_BLOCK_INDENT, strlen (HTML_BLOCK_INDENT)),
-			      text_width (painter, desc, HTML_BLOCK_CITE, strlen (HTML_BLOCK_CITE)));
-}
-
-static void
-html_painter_real_ref_font (HTMLPainter *painter, HTMLFont *font)
-{
-}
-
-static void
-html_painter_real_unref_font (HTMLPainter *painter, HTMLFont *font)
-{
-	if (font->ref_count < 1) {
-		pango_font_description_free (font->data);
-		font->data = NULL;
-	}
-}
-
 static void
 html_painter_class_init (GObjectClass *object_class)
 {
@@ -278,10 +231,6 @@ html_painter_class_init (GObjectClass *object_class)
 	class->begin = (gpointer) begin_unimplemented;
 	class->end = (gpointer) end_unimplemented;
 
-	class->alloc_font = html_painter_real_alloc_font;
-	class->ref_font   = html_painter_real_ref_font;
-	class->unref_font = html_painter_real_unref_font;
-
 	class->alloc_color = (gpointer) alloc_color_unimplemented;
 	class->free_color = (gpointer) free_color_unimplemented;
 
@@ -289,7 +238,7 @@ html_painter_class_init (GObjectClass *object_class)
 	class->get_black = (gpointer) get_black_unimplemented;
 	class->draw_line = (gpointer) draw_line_unimplemented;
 	class->draw_rect = (gpointer) draw_rect_unimplemented;
-	class->draw_text = (gpointer) draw_text_unimplemented;
+	class->draw_glyphs = (gpointer) draw_glyphs_unimplemented;
 	class->draw_spell_error = (gpointer) draw_spell_error_unimplemented;
 	class->fill_rect = (gpointer) fill_rect_unimplemented;
 	class->draw_pixmap = (gpointer) draw_pixmap_unimplemented;
@@ -297,7 +246,7 @@ html_painter_class_init (GObjectClass *object_class)
 	class->clear = (gpointer) clear_unimplemented;
 	class->set_background_color = (gpointer) set_background_color_unimplemented;
 	class->draw_shade_line = (gpointer) draw_shade_line_unimplemented;
-	class->draw_panel = (gpointer) draw_panel_unimplemented;
+	class->draw_border = (gpointer) draw_border_unimplemented;
 
 	class->set_clip_rectangle = (gpointer) set_clip_rectangle_unimplemented;
 	class->draw_background = (gpointer) draw_background_unimplemented;
@@ -341,6 +290,8 @@ html_painter_begin (HTMLPainter *painter,
 	g_return_if_fail (painter != NULL);
 	g_return_if_fail (HTML_IS_PAINTER (painter));
 
+	painter->clip_height = painter->clip_width = 0;
+
 	(* HP_CLASS (painter)->begin) (painter, x1, y1, x2, y2);
 }
 
@@ -422,52 +373,115 @@ html_painter_get_font (HTMLPainter *painter, HTMLFontFace *face, GtkHTMLFontStyl
 	return font ? font->data : NULL;
 }
 
-void
-html_painter_calc_text_size (HTMLPainter *painter,
-			     const gchar *text,
-			     guint len, HTMLTextPangoInfo *pi, PangoAttrList *attrs, GList *glyphs, gint start_byte_offset, gint *line_offset,
-			     GtkHTMLFontStyle font_style,
-			     HTMLFontFace *face,
-			     gint *width, gint *asc, gint *dsc)
+static void
+get_font_info (HTMLPainter       *painter,
+	       HTMLTextPangoInfo *pi,
+	       HTMLFontFace     **font_face,
+	       GtkHTMLFontStyle  *font_style)
 {
+	if (pi && pi->have_font) {
+		*font_face = pi->face;
+		*font_style = pi->font_style;
+	} else {
+		*font_face = painter->font_face;
+		*font_style = painter->font_style;
+	}
+}
+	   
+
+static int
+get_space_width (HTMLPainter       *painter,
+		 HTMLTextPangoInfo *pi)
+{
+	HTMLFontFace    *font_face;
+	GtkHTMLFontStyle font_style;
+
+	get_font_info (painter, pi, &font_face, &font_style);
+	
+	return html_painter_get_space_width (painter, font_style, font_face);
+}
+
+/**
+ * html_painter_calc_entries_size:
+ * @painter: a #HTMLPainter
+ * @text: text to compute size of
+ * @len: length of text, in characters
+ * @pi: #HTMLTextPangoInfo structure holding information about
+ *   the text. (It may be for a larger range of text that includes
+ *   the range specified by @text and @len.)
+ * @glyphs: list holding information about segments of text to draw.
+ *   There is one segment for each run of non-tab characters. The
+ *   list is structure as a series of pairs of PangoGlyphString,
+ *   GINT_TO_POINTER(item_index), where item_index is an index
+ *   within pi->entries[].
+ * @line_offset: location to store column offset in output, used
+ *   for tab display. On entry holds the the column offset for the
+ *   first character in @text. On exit, updated to hold the character
+ *   offset for the first character after the end of the text.
+ *   If %NULL is passed, then tabs are disabled and substituted with spaces.
+ * @width: location to store width of text (in engine coordinates)
+ * @asc: location to store ascent of text (in engine coordinates)
+ * @dsc: location to store descent of text (in engine coordinates)
+ * 
+ * Computes size information for a piece of text, using provided Pango
+ * layout information.
+ **/
+void
+html_painter_calc_entries_size (HTMLPainter *painter,
+				const gchar *text, guint len,
+				HTMLTextPangoInfo *pi, GList *glyphs, gint *line_offset,
+				gint *width, gint *asc, gint *dsc)
+{
+	HTMLFontFace    *font_face;
+	GtkHTMLFontStyle font_style;
 	HTMLFont *font;
 	
 	g_return_if_fail (painter != NULL);
 	g_return_if_fail (HTML_IS_PAINTER (painter));
 	g_return_if_fail (text != NULL);
 
-	font = html_font_manager_get_font (&painter->font_manager, face, font_style);
+	if (line_offset || !pi) {
+		get_font_info (painter, pi, &font_face, &font_style);
+		font = html_font_manager_get_font (&painter->font_manager, font_face, font_style);
+	} else
+		font = NULL;
 
 	text_size (painter, (PangoFontDescription *) font->data, text, g_utf8_offset_to_pointer (text, len) - text,
-		   pi, attrs, glyphs, start_byte_offset, width, asc, dsc);
+		   pi, glyphs, width, asc, dsc);
 	/* g_print ("calc_text_size %s %d %d %d\n", text, *width, asc ? *asc : -1, dsc ? *dsc : -1); */
 
 	if (line_offset) {
+		gint space_width = html_painter_get_space_width (painter, font_style, font_face);
 		gint tabs;
-		*width += (html_text_text_line_length (text, line_offset, len, &tabs) - len + tabs)*html_painter_get_space_width (painter, font_style, face);
+		
+		*width += (html_text_text_line_length (text, line_offset, len, &tabs) - len + tabs)*space_width;
 	}
 }
 
+/**
+ * html_painter_calc_text_size:
+ * @painter: a #HTMLPainter
+ * @text: text to compute size of
+ * @len: length of text, in characters
+ * @width: location to store width of text (in engine coordinates)
+ * @asc: location to store ascent of text (in engine coordinates)
+ * @dsc: location to store descent of text (in engine coordinates)
+ * 
+ * Computes size information for a piece of text.
+ **/
 void
-html_painter_calc_text_size_bytes (HTMLPainter *painter,
-				   const gchar *text,
-				   guint bytes_len, HTMLTextPangoInfo *pi, PangoAttrList *attrs, GList *glyphs, gint start_byte_offset, gint *line_offset,
-				   HTMLFont *font, GtkHTMLFontStyle style,
-				   gint *width, gint *asc, gint *dsc)
+html_painter_calc_text_size (HTMLPainter *painter,
+			     const gchar *text, guint len,
+			     gint *width, gint *asc, gint *dsc)
 {
+	gint line_offset = 0;
+	
 	g_return_if_fail (painter != NULL);
 	g_return_if_fail (HTML_IS_PAINTER (painter));
 	g_return_if_fail (text != NULL);
-	g_return_if_fail (style != GTK_HTML_FONT_STYLE_DEFAULT);
 
-	text_size (painter, (PangoFontDescription *) font->data, text, bytes_len, pi, attrs, glyphs, start_byte_offset, width, asc, dsc);
-	
-	/* g_print ("calc_text_size_bytes %d %d %d\n", *width, asc ? *asc : -1, dsc ? *dsc : -1); */
-	
-	if (line_offset) {
-		gint tabs, len = g_utf8_pointer_to_offset (text, text + bytes_len);
-		*width += (html_text_text_line_length (text, line_offset, len, &tabs) - len + tabs)*font->space_width;
-	}
+	html_painter_calc_entries_size (painter, text, len, NULL, NULL, &line_offset,
+					width, asc, dsc);
 }
 
 /* The actual paint operations.  */
@@ -527,68 +541,137 @@ html_replace_tabs (const gchar *text, gchar *translated, guint bytes)
 	} while (tab);
 }
 
-static inline GList *
-shift_items (GList *items, gint byte_offset)
+/**
+ * html_painter_draw_entries:
+ * @painter: a #HTMLPainter
+ * @x: x coordinate at which to draw text, in engine coordinates
+ * @y: x coordinate at which to draw text, in engine coordinates
+ * @text: text to draw
+ * @len: length of text, in characters
+ * @pi: #HTMLTextPangoInfo structure holding information about
+ *   the text. (It may be for a larger range of text that includes
+ *   the range specified by @text and @len.)
+ * @glyphs: list holding information about segments of text to draw.
+ *   There is one segment for each run of non-tab characters. The
+ *   list is structure as a series of pairs of PangoGlyphString,
+ *   GINT_TO_POINTER(item_index), where item_index is an index
+ *   within pi->entries[].
+ * @line_offset: column offset of the first character in @text, used for
+ *  tab display. If set to -1, then tabs are disabled and substituted
+ * with spaces.
+ * 
+ * Draws a piece of text, using provided Pango layout information.
+ **/
+void
+html_painter_draw_entries (HTMLPainter *painter, gint x, gint y,
+			   const gchar *text, gint len,
+			   HTMLTextPangoInfo *pi, GList *glyphs,
+			   gint line_offset)
 {
-	if (items) {
-		PangoItem *item;
-
-		while (items && (item = (PangoItem *) items->data) && item->offset + item->length <= byte_offset)
-			items = items->next;
-	}
-
-	return items;
-}
-
-static inline GList *
-shift_glyphs (GList *glyphs, gint len)
-{
-	if (glyphs) {
-		PangoGlyphString *str;
-
-		while (glyphs && (str = (PangoGlyphString *) glyphs->data) && len > 0) {
-			len -= str->num_glyphs;
-			glyphs = glyphs->next->next;
-		}
-	}
-
-	return glyphs;
-}
-
-gint
-html_painter_draw_text (HTMLPainter *painter, gint x, gint y,
-			const gchar *text, gint len, HTMLTextPangoInfo *pi, PangoAttrList *attrs, GList *glyphs, gint start_byte_offset, gint line_offset)
-{
-	const gchar *tab, *c_text = text;
-	gint bytes, byte_offset = 0;
-
-	g_return_val_if_fail (painter != NULL, line_offset);
-	g_return_val_if_fail (HTML_IS_PAINTER (painter), line_offset);
+	const gchar *tab, *c_text;
+	gint bytes;
+	GList *gl;
+	int first_item_offset = -1;
+	int space_width = -1;
+	
+	g_return_if_fail (painter != NULL);
+	g_return_if_fail (HTML_IS_PAINTER (painter));
 
+	c_text = text;
 	bytes = g_utf8_offset_to_pointer (text, len) - text;
-	while ((tab = memchr (c_text, (unsigned char) '\t', bytes))) {
-		gint c_bytes = tab - c_text;
-		gint c_len = g_utf8_pointer_to_offset (c_text, tab);
+	gl = glyphs;
+	tab = memchr (c_text, (unsigned char) '\t', bytes);
+
+	/* We iterate through the string by tabs and non-tab segments, skipping over
+	 * the tabs and drawing the non-tab segments.
+	 *
+	 * We have one PangoGlyphString for each non-tab segment within each item.
+	 */
+	while (gl) {
+		gint ii = GPOINTER_TO_INT (gl->next->data);
+		PangoItem *item = pi->entries[ii].glyph_item.item;
+		const gchar *item_end;
+		const gchar *next;
+
+		if (first_item_offset < 0)
+			first_item_offset = item->offset;
+
+		item_end = text + item->offset - first_item_offset + item->length;
 		
-		if (c_bytes)
-			x += (* HP_CLASS (painter)->draw_text) (painter, x, y, c_text, c_len, pi, NULL, glyphs, start_byte_offset + (c_text - text));
-		if (line_offset == -1)
-			x += html_painter_get_space_width (painter, painter->font_style, painter->font_face);
-		else {
-			line_offset += c_len;
-			x += html_painter_get_space_width (painter, painter->font_style, painter->font_face)*(8 - (line_offset % 8));
-			line_offset += 8 - (line_offset % 8);
+		if (*c_text == '\t')
+			next = c_text + 1;
+		else if (tab && tab < item_end)
+			next = tab;
+		else
+			next = item_end;
+
+		if (*c_text == '\t') {
+			if (space_width < 0)
+				space_width = get_space_width (painter, pi);
+
+			if (line_offset == -1)
+				x += space_width;
+			else {
+				x += space_width * (8 - (line_offset % 8));
+				line_offset += 8 - (line_offset % 8);
+			}
+			
+			tab = memchr (c_text + 1, (unsigned char) '\t', bytes - 1);
+		} else {
+			x += (* HP_CLASS (painter)->draw_glyphs) (painter, x, y, item, gl->data, NULL, NULL);
+
+			if (line_offset != -1)
+				line_offset += g_utf8_pointer_to_offset (c_text, next);
+			
+			gl = gl->next->next;
 		}
-		c_text += c_bytes + 1;
-		bytes -= c_bytes + 1;
-		byte_offset += c_bytes + 1;
-		glyphs = shift_glyphs (glyphs, c_len);
-		len -= c_len + 1;
+		
+		bytes -= next - c_text;
+		c_text = next;
 	}
+}
 
-	(* HP_CLASS (painter)->draw_text) (painter, x, y, c_text, len, pi, attrs, glyphs, start_byte_offset + (c_text - text));
+int
+html_painter_draw_glyphs (HTMLPainter *painter, int x, int y, PangoItem *item, PangoGlyphString *glyphs, GdkColor *fg, GdkColor *bg)
+{
+	return (* HP_CLASS (painter)->draw_glyphs) (painter, x, y, item, glyphs, fg, bg);
+}
 
-	return line_offset + len;
+/**
+ * html_painter_draw_text:
+ * @painter: a #HTMLPainter
+ * @x: x coordinate at which to draw text, in engine coordinates
+ * @y: x coordinate at which to draw text, in engine coordinates
+ * @text: text to draw
+ * @len: length of text, in characters
+ * 
+ * Draws a piece of text.
+ **/
+void
+html_painter_draw_text (HTMLPainter *painter, gint x, gint y,
+			const gchar *text, gint len)
+{
+	HTMLTextPangoInfo *pi;
+	GList *glyphs;
+	gint blen;
+
+	g_return_if_fail (painter != NULL);
+	g_return_if_fail (HTML_IS_PAINTER (painter));
+
+	if (len < 0)
+		len = g_utf8_strlen (text, -1);
+
+	blen = g_utf8_offset_to_pointer (text, len) - text;
+
+	pi = html_painter_text_itemize_and_prepare_glyphs (painter, html_painter_get_font (painter, painter->font_face, painter->font_style),
+							   text, blen, &glyphs, NULL);
+
+	html_painter_draw_entries (painter, x, y, text, len, pi, glyphs, 0);
+	
+	if (glyphs)
+		html_painter_glyphs_destroy (glyphs);
+	if (pi)
+		html_text_pango_info_destroy (pi);
 }
 
 void
@@ -659,17 +742,17 @@ html_painter_draw_shade_line (HTMLPainter *painter,
 }
 
 void
-html_painter_draw_panel (HTMLPainter *painter,
-			 GdkColor *bg,
-			 gint x, gint y,
-			 gint width, gint height,
-			 GtkHTMLEtchStyle inset,
-			 gint bordersize)
+html_painter_draw_border (HTMLPainter *painter,
+			  GdkColor *bg,
+			  gint x, gint y,
+			  gint width, gint height,
+			  HTMLBorderStyle style,
+			  gint bordersize)
 {
 	g_return_if_fail (painter != NULL);
 	g_return_if_fail (HTML_IS_PAINTER (painter));
 
-	(* HP_CLASS (painter)->draw_panel) (painter, bg, x, y, width, height, inset, bordersize);
+	(* HP_CLASS (painter)->draw_border) (painter, bg, x, y, width, height, style, bordersize);
 }
 
 void  
@@ -691,9 +774,27 @@ html_painter_set_clip_rectangle (HTMLPainter *painter,
 	g_return_if_fail (painter != NULL);
 	g_return_if_fail (HTML_IS_PAINTER (painter));
 
+	painter->clip_x = x;
+	painter->clip_y = y;
+	painter->clip_width = width;
+	painter->clip_height = height;
+
+	/* printf ("clip rect: %d,%d %dx%d\n", x, y, width, height); */
+
 	(* HP_CLASS (painter)->set_clip_rectangle) (painter, x, y, width, height);
 }
 
+void
+html_painter_get_clip_rectangle (HTMLPainter *painter,
+				 gint *x, gint *y,
+				 gint *width, gint *height)
+{
+	*x = painter->clip_x;
+	*y = painter->clip_y;
+	*width = painter->clip_width;
+	*height = painter->clip_height;
+}
+
 /* Passing 0 for pix_width / pix_height makes it use the image width */
 void
 html_painter_draw_background (HTMLPainter *painter,
@@ -721,28 +822,57 @@ html_painter_get_pixel_size (HTMLPainter *painter)
 gint
 html_painter_draw_spell_error (HTMLPainter *painter,
 			       gint x, gint y,
-			       const gchar *text,
-			       gint len, HTMLTextPangoInfo *pi, GList *glyphs, gint start_byte_offset)
+			       HTMLTextPangoInfo *pi, GList *glyphs)
 {
-	return (* HP_CLASS (painter)->draw_spell_error) (painter, x, y, text, len, pi, glyphs, start_byte_offset);
+	return (* HP_CLASS (painter)->draw_spell_error) (painter, x, y, pi, glyphs);
 }
 
 HTMLFont *
-html_painter_alloc_font (HTMLPainter *painter, gchar *face_name, gdouble size, gboolean points, GtkHTMLFontStyle style)
+html_painter_alloc_font (HTMLPainter *painter, gchar *face, gdouble size, gboolean points, GtkHTMLFontStyle style)
 {
-	return (* HP_CLASS (painter)->alloc_font) (painter, face_name, size, points, style);
+	PangoFontDescription *desc = NULL;
+	gint space_width, space_asc, space_dsc;
+
+	if (face) {
+		desc = pango_font_description_from_string (face);
+		pango_font_description_set_size (desc, (gint) size);
+	}
+
+	if (!desc || !pango_font_description_get_family (desc)) {
+		if (desc)
+			pango_font_description_free (desc);
+
+		desc = pango_font_description_copy (gtk_widget_get_style (painter->widget)->font_desc);
+	}
+
+	pango_font_description_set_size (desc, size);
+	pango_font_description_set_style (desc, style & GTK_HTML_FONT_STYLE_ITALIC ? PANGO_STYLE_ITALIC : PANGO_STYLE_NORMAL);
+	pango_font_description_set_weight (desc, style & GTK_HTML_FONT_STYLE_BOLD ? PANGO_WEIGHT_BOLD : PANGO_WEIGHT_NORMAL);
+
+	text_size (painter, desc, " ", 1, NULL, NULL, &space_width, &space_asc, &space_dsc);
+
+	return html_font_new (desc,
+			      space_width,
+			      space_asc, space_dsc,
+			      text_width (painter, desc, "\xc2\xa0", 2),
+			      text_width (painter, desc, "\t", 1),
+			      text_width (painter, desc, "e", 1),
+			      text_width (painter, desc, HTML_BLOCK_INDENT, strlen (HTML_BLOCK_INDENT)),
+			      text_width (painter, desc, HTML_BLOCK_CITE, strlen (HTML_BLOCK_CITE)));
 }
 
 void
 html_painter_ref_font (HTMLPainter *painter, HTMLFont *font)
 {
-	(* HP_CLASS (painter)->ref_font) (painter, font);
 }
 
 void
 html_painter_unref_font (HTMLPainter *painter, HTMLFont *font)
 {
-	(* HP_CLASS (painter)->unref_font) (painter, font);
+	if (font->ref_count < 1) {
+		pango_font_description_free (font->data);
+		font->data = NULL;
+	}
 }
 
 guint
@@ -808,7 +938,8 @@ gint
 html_painter_pango_to_engine (HTMLPainter       *painter,
 			      gint               pango_units)
 {
-	return (gint)(0.5 + pango_units / painter->engine_to_pango);
+	gdouble tmp = 0.5 + pango_units / painter->engine_to_pango;
+	return (gint) CLAMP(tmp, G_MININT, G_MAXINT);
 }
 
 /**
@@ -826,7 +957,8 @@ gint
 html_painter_engine_to_pango (HTMLPainter       *painter,
 			      gint               engine_units)
 {
-	return (gint)(0.5 + engine_units * painter->engine_to_pango);
+	gdouble tmp = 0.5 + engine_units * painter->engine_to_pango;
+	return (gint) CLAMP(tmp, G_MININT, G_MAXINT);
 }
 
 void
@@ -851,6 +983,8 @@ html_painter_text_itemize_and_prepare_glyphs (HTMLPainter *painter, PangoFontDes
 
 	/* printf ("itemize + glyphs\n"); */
 
+	*glyphs = NULL;
+
 	if (empty_attrs) {
 		attrs = pango_attr_list_new ();
 		attr = pango_attr_font_desc_new (desc);
@@ -858,7 +992,9 @@ html_painter_text_itemize_and_prepare_glyphs (HTMLPainter *painter, PangoFontDes
 		attr->end_index = bytes;
 		pango_attr_list_insert (attrs, attr);
 	}
+
 	items = pango_itemize (painter->pango_context, text, 0, bytes, attrs, NULL);
+
 	if (empty_attrs)
 		pango_attr_list_unref (attrs);
 
@@ -870,10 +1006,9 @@ html_painter_text_itemize_and_prepare_glyphs (HTMLPainter *painter, PangoFontDes
 
 		pi = html_text_pango_info_new (g_list_length (items));
 
-		*glyphs = NULL;
 		for (il = items; il; il = il->next) {
 			item = (PangoItem *) il->data;
-			pi->entries [i].item = item;
+			pi->entries [i].glyph_item.item = item;
 			end = g_utf8_offset_to_pointer (text, item->num_chars);
 			*glyphs = html_get_glyphs_non_tab (*glyphs, item, i, text, end - text, item->num_chars);
 			text = end;
@@ -881,8 +1016,7 @@ html_painter_text_itemize_and_prepare_glyphs (HTMLPainter *painter, PangoFontDes
 		}
 		*glyphs = g_list_reverse (*glyphs);
 		g_list_free (items);
-	} else
-		*glyphs = NULL;
+	}
 
 	return pi;
 }
